import { TypedEventEmitter } from "@libp2p/interface";
import { Logger } from "@libp2p/logger";
import { Node, Tree, ReadWriteTransaction } from "@canvas-js/okra";
import type { Signature, Signer, Message, Awaitable } from "@canvas-js/interfaces";
import type { AbstractModelDB } from "@canvas-js/modeldb";
import type { SyncServer } from "./interface.js";
import { SignedMessage } from "./SignedMessage.js";
export type GossipLogConsumer<Payload = unknown> = (this: AbstractGossipLog<Payload>, { id, signature, message }: SignedMessage<Payload>) => Awaitable<void>;
export interface GossipLogInit<Payload = unknown> {
    topic: string;
    apply: GossipLogConsumer<Payload>;
    validatePayload?: (payload: unknown) => payload is Payload;
    verifySignature?: (signature: Signature, message: Message<Payload>) => Awaitable<void>;
    signer?: Signer<Payload>;
    indexAncestors?: boolean;
    rebuildMerkleIndex?: boolean;
}
export type GossipLogEvents<Payload = unknown> = {
    message: CustomEvent<{
        id: string;
        signature: Signature;
        message: Message<Payload>;
    }>;
    commit: CustomEvent<{
        root: Node;
    }>;
    sync: CustomEvent<{
        peerId?: string;
        duration: number;
        messageCount: number;
    }>;
    error: CustomEvent<{
        error: Error;
    }>;
};
export declare abstract class AbstractGossipLog<Payload = unknown> extends TypedEventEmitter<GossipLogEvents<Payload>> {
    #private;
    static schema: {
        $ancestors: {
            id: "primary";
            links: "json";
        };
        $messages: {
            id: "primary";
            signature: "json";
            message: "json";
            hash: "string";
        };
        $heads: {
            id: "primary";
        };
    };
    readonly topic: string;
    readonly indexAncestors: boolean;
    readonly signer: Signer<Payload>;
    abstract db: AbstractModelDB;
    abstract tree: Tree;
    abstract close(): Promise<void>;
    protected readonly log: Logger;
    readonly validatePayload: (payload: unknown) => payload is Payload;
    readonly verifySignature: (signature: Signature, message: Message<Payload>) => Awaitable<void>;
    protected constructor(init: GossipLogInit<Payload>);
    replay(): Promise<void>;
    encode(signature: Signature, message: Message<Payload>): SignedMessage<Payload>;
    decode(value: Uint8Array): SignedMessage<Payload>;
    getClock(): Promise<[clock: number, heads: string[]]>;
    has(id: string): Promise<boolean>;
    get(id: string): Promise<[signature: Signature, message: Message<Payload>] | [null, null]>;
    export(range?: {
        lt?: string;
        lte?: string;
        gt?: string;
        gte?: string;
        reverse?: boolean;
        limit?: number;
    }): Promise<{
        id: string;
        signature: Signature;
        message: Message<Payload>;
    }[]>;
    /**
     * Sign and append a new *unsigned* message to the end of the log.
     * The concurrent heads of the local log are used as parents.
     */
    append(payload: Payload, options?: {
        signer?: Signer<Payload>;
    }): Promise<SignedMessage<Payload>>;
    getMissingParents(txn: ReadWriteTransaction, parents: string[]): Set<string>;
    /**
     * Insert an existing signed message into the log (ie received via HTTP API).
     * If any of the parents are not present, throw an error.
     */
    insert(signedMessage: SignedMessage<Payload>): Promise<{
        id: string;
    }>;
    private apply;
    getAncestors(id: string, atOrBefore: number): Promise<string[]>;
    isAncestor(id: string, ancestor: string, visited?: Set<string>): Promise<boolean>;
    /**
     * Sync with a remote source, applying and inserting all missing messages into the local log
     */
    sync(server: SyncServer, options?: {
        sourceId?: string;
    }): Promise<{
        messageCount: number;
    }>;
    serve<T>(callback: (txn: SyncServer) => Awaitable<T>): Promise<T>;
}
