var __classPrivateFieldSet = (this && this.__classPrivateFieldSet) || function (receiver, state, value, kind, f) {
    if (kind === "m") throw new TypeError("Private method is not writable");
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a setter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot write private member to an object whose class did not declare it");
    return (kind === "a" ? f.call(receiver, value) : f ? f.value = value : state.set(receiver, value)), value;
};
var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, state, kind, f) {
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a getter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot read private member from an object whose class did not declare it");
    return kind === "m" ? f : kind === "a" ? f.call(receiver) : f ? f.value : state.get(receiver);
};
var _AbstractGossipLog_apply;
import { TypedEventEmitter, CustomEvent } from "@libp2p/interface";
import { logger } from "@libp2p/logger";
import { equals, toString } from "uint8arrays";
import { hashEntry } from "@canvas-js/okra";
import { ed25519 } from "@canvas-js/signatures";
import { assert, zip } from "@canvas-js/utils";
import { Driver } from "./sync/driver.js";
import { AncestorIndex } from "./AncestorIndex.js";
import { SignedMessage } from "./SignedMessage.js";
import { decodeId, encodeId, messageIdPattern } from "./ids.js";
import { getNextClock } from "./schema.js";
import { topicPattern } from "./utils.js";
export class AbstractGossipLog extends TypedEventEmitter {
    constructor(init) {
        super();
        _AbstractGossipLog_apply.set(this, void 0);
        assert(topicPattern.test(init.topic), "invalid topic (must match [a-zA-Z0-9\\.\\-])");
        this.topic = init.topic;
        this.indexAncestors = init.indexAncestors ?? false;
        this.signer = init.signer ?? ed25519.create();
        __classPrivateFieldSet(this, _AbstractGossipLog_apply, init.apply, "f");
        this.validatePayload = init.validatePayload ?? ((payload) => true);
        this.verifySignature = init.verifySignature ?? this.signer.scheme.verify;
        this.log = logger(`canvas:gossiplog:[${this.topic}]`);
    }
    async replay() {
        await this.tree.read(async (txn) => {
            for (const leaf of txn.keys()) {
                const id = decodeId(leaf);
                const record = await this.db.get("$messages", id);
                if (record === null) {
                    this.log.error("failed to get message %s from database", id);
                    continue;
                }
                const signedMessage = this.encode(record.signature, record.message);
                assert(signedMessage.id === id);
                await __classPrivateFieldGet(this, _AbstractGossipLog_apply, "f").apply(this, [signedMessage]);
            }
        });
    }
    encode(signature, message) {
        assert(this.topic === message.topic, "expected this.topic === topic");
        assert(this.validatePayload(message.payload), "error encoding message (invalid payload)");
        return SignedMessage.encode(signature, message);
    }
    decode(value) {
        const signedMessage = SignedMessage.decode(value);
        assert(this.topic === signedMessage.message.topic, "expected this.topic === topic");
        assert(this.validatePayload(signedMessage.message.payload), "error decoding message (invalid payload)");
        return signedMessage;
    }
    async getClock() {
        const heads = await this.db.query("$heads", { orderBy: { id: "asc" } });
        const ids = heads.map(({ id }) => id);
        const clock = getNextClock(ids.map(encodeId));
        return [clock, ids];
    }
    async has(id) {
        assert(messageIdPattern.test(id), "invalid message ID");
        const records = await this.db.query("$messages", { select: { id: true }, where: { id } });
        return records.length > 0;
    }
    async get(id) {
        const record = await this.db.get("$messages", id);
        if (record === null) {
            return [null, null];
        }
        else {
            return [record.signature, record.message];
        }
    }
    export(range = {}) {
        const { reverse = false, limit, ...where } = range;
        return this.db.query("$messages", {
            where: { id: where },
            select: { id: true, signature: true, message: true },
            orderBy: { id: reverse ? "desc" : "asc" },
            limit,
        });
    }
    /**
     * Sign and append a new *unsigned* message to the end of the log.
     * The concurrent heads of the local log are used as parents.
     */
    async append(payload, options = {}) {
        const signer = options.signer ?? this.signer;
        let root = null;
        const signedMessage = await this.tree.write(async (txn) => {
            const [clock, parents] = await this.getClock();
            const message = { topic: this.topic, clock, parents, payload };
            const signature = await signer.sign(message);
            const signedMessage = this.encode(signature, message);
            this.log("appending message %s", signedMessage.id);
            await this.apply(txn, signedMessage);
            root = txn.getRoot();
            return signedMessage;
        });
        assert(root !== null, "failed to commit transaction");
        this.dispatchEvent(new CustomEvent("commit", { detail: { root } }));
        return signedMessage;
    }
    getMissingParents(txn, parents) {
        const missingParents = new Set();
        this.log("looking up %s parents", parents.length);
        for (const parentId of parents) {
            // TODO: txn.messages.getMany
            const leaf = txn.getNode(0, encodeId(parentId));
            if (leaf !== null) {
                this.log("found parent %s", parentId);
            }
            else {
                this.log("missing parent %s", parentId);
                missingParents.add(parentId);
            }
        }
        return missingParents;
    }
    /**
     * Insert an existing signed message into the log (ie received via HTTP API).
     * If any of the parents are not present, throw an error.
     */
    async insert(signedMessage) {
        this.log("inserting message %s", signedMessage.id);
        await this.verifySignature(signedMessage.signature, signedMessage.message);
        const parentKeys = signedMessage.message.parents.map(encodeId);
        let root = null;
        await this.tree.write(async (txn) => {
            const hasSignedMessage = await this.has(signedMessage.id);
            if (hasSignedMessage) {
                return;
            }
            for (const parentKey of parentKeys) {
                const leaf = txn.getNode(0, parentKey);
                if (leaf === null) {
                    this.log.error("missing parent %s of message %s: %O", parent, signedMessage.id, signedMessage.message);
                    throw new Error(`missing parent ${parent} of message ${signedMessage.id}`);
                }
            }
            await this.apply(txn, signedMessage);
            root = txn.getRoot();
        });
        assert(root !== null, "failed to commit transaction");
        this.dispatchEvent(new CustomEvent("commit", { detail: { root } }));
        return { id: signedMessage.id };
    }
    async apply(txn, signedMessage) {
        this.log("applying %s %O", signedMessage.id, signedMessage.message);
        try {
            await __classPrivateFieldGet(this, _AbstractGossipLog_apply, "f").apply(this, [signedMessage]);
        }
        catch (error) {
            this.dispatchEvent(new CustomEvent("error", { detail: { error } }));
            throw error;
        }
        const { id, signature, message, key, value } = signedMessage;
        const hash = toString(hashEntry(key, value), "hex");
        await this.db.set("$messages", { id, signature, message, hash });
        const heads = await this.db.query("$heads");
        await this.db.apply([
            ...heads
                .filter((head) => message.parents.includes(head.id))
                .map((head) => ({ model: "$heads", operation: "delete", key: head.id })),
            { model: "$heads", operation: "set", value: { id } },
        ]);
        txn.set(key, value);
        // const root = await txn.getRoot()
        // await new MerkleIndex(this.db).commit(root)
        await new AncestorIndex(this.db).indexAncestors(id, message.parents);
        this.dispatchEvent(new CustomEvent("message", { detail: { id, signature, message } }));
    }
    async getAncestors(id, atOrBefore) {
        const results = await new AncestorIndex(this.db).getAncestors(id, atOrBefore);
        this.log("getAncestors of %s atOrBefore %d: %o", id, atOrBefore, results);
        return Array.from(results).sort();
    }
    async isAncestor(id, ancestor, visited = new Set()) {
        return await new AncestorIndex(this.db).isAncestor(id, ancestor, visited);
    }
    /**
     * Sync with a remote source, applying and inserting all missing messages into the local log
     */
    async sync(server, options = {}) {
        const start = performance.now();
        let messageCount = 0;
        await this.tree.read(async (txn) => {
            const driver = new Driver(this.topic, server, txn);
            for await (const keys of driver.sync()) {
                const values = await server.getValues(keys);
                for (const [key, value] of zip(keys, values)) {
                    const signedMessage = this.decode(value);
                    assert(equals(key, signedMessage.key), "invalid message key");
                    await this.verifySignature(signedMessage.signature, signedMessage.message);
                    await this.insert(signedMessage);
                    messageCount++;
                }
            }
        });
        const duration = Math.ceil(performance.now() - start);
        this.log("finished sync with peer %s (%d messages in %dms)", options.sourceId, messageCount, duration);
        this.dispatchEvent(new CustomEvent("sync", { detail: { peerId: options.sourceId, messageCount, duration } }));
        return { messageCount };
    }
    serve(callback) {
        return this.tree.read((txn) => callback({
            getRoot: () => txn.getRoot(),
            getNode: (level, key) => txn.getNode(level, key),
            getChildren: (level, key) => txn.getChildren(level, key),
            getValues: async (keys) => {
                const values = [];
                for (const key of keys) {
                    const [signature, message] = await this.get(decodeId(key));
                    if (signature === null || message === null) {
                        throw new Error("wtf??");
                    }
                    const signedMessage = SignedMessage.encode(signature, message);
                    assert(equals(signedMessage.key, key), "invalid message key");
                    values.push(signedMessage.value);
                }
                return values;
            },
        }));
    }
}
_AbstractGossipLog_apply = new WeakMap();
AbstractGossipLog.schema = {
    $messages: { id: "primary", signature: "json", message: "json", hash: "string" },
    $heads: { id: "primary" },
    ...AncestorIndex.schema,
};
